/* SoriBada - compatible client with windows soribada
 * Copyright (C) 2002 Choe Hwanjin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>

#include "mp3.h"
#include "genre.h"

static const char *version_tbl[3] = {
	"1.0", "2.0", "2.5"
};

static const char *layer_tbl[3] = {
	"I", "II", "III"
};

static const unsigned int bitrate_tbl[3][3][15] = {
{
	{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448},
	{0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384},
	{0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320}
},
{
	/* somebody says this is not right
	{0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
	{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
	{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
	*/
	{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448},
	{0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384},
	{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
},
{
	/* somebody says this is not right
	{0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
	{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
	{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
	*/
	{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448},
	{0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384},
	{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
}
};

static const unsigned int frequency_tbl[3][4] = {
	{44100, 48000, 32000, 0},
	{22050, 24000, 16000, 0},
	{11025, 12000,  8000, 0}
};

static const char *mode_tbl[4] = {
	"Stereo", "Joint stereo", "Dual channel", "Mono"
};

const char *mp3_get_version_str(struct mp3info *info) {
	return (const char *)version_tbl[info->version];
}

const char *mp3_get_layer_str(struct mp3info *info) {
	return (const char *)layer_tbl[info->layer];
}

const char *mp3_get_mode_str(struct mp3info *info) {
	return (const char *)mode_tbl[info->mode];
}

const char *mp3_get_genre_str(struct mp3info *info) {
	return (const char *)genre_tbl[info->mode];
}

void copy_string(char *dest, const char *src, int n) {
	int length = strlen(src);
	memcpy(dest, src, length > n ? n : length);
}

static long convert_to_long(unsigned char *buf) {
	return (buf[3] << 24) | (buf[2] << 16) | (buf[1] << 8) | buf[0];
}

/*
static unsigned long convert_to_magic(unsigned char *buf) {
	return (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
}
*/

int is_valid_header(unsigned char *header) {
	/*
	printf("MAGIC: 0x%lx\n", convert_to_magic(header));
	*/
	if (header[0] != 0xff || (header[1] & 0xe0) != 0xe0)	/* Frame sync */
		return FALSE;
	if ((header[1] & 0x18) == 0x08)				/* version */
		return FALSE;
	if ((header[1] & 0x06) == 0x00)				/* Layer */
		return FALSE;
	if ((header[2] & 0xf0) == 0xf0)				/* bitrate */
		return FALSE;
	if ((header[2] & 0x0c) == 0x0c)				/* frequency */
		return FALSE;

	return TRUE;
}

int checkheader(FILE *file) {
	unsigned char header[8];
	int read, i;
	long offset = 0;

	/* check for WAV format or Wrong RIFF header */
	fread(header, 8, 1, file);
	if (header[0] == 'R' && header[1] == 'I' &&\
	    header[2] == 'F' && header[3] == 'F') {
		/* found wave RIFF header */
		fread(header, 8, 1, file);
		if (header[0] == 'W' && header[1] == 'A' &&\
		    header[2] == 'V' && header[3] == 'E' &&\
		    header[4] == 'f' && header[5] == 'm' &&\
		    header[6] == 't' && header[7] == ' ') {
			/* get rid of WAVE header */
			fread(header, 4, 1, file);
			offset = convert_to_long(header);
			fseek(file, offset, SEEK_CUR);
			for (i = 0; i < 2048; i++) {
				read = fread(header, 4, 1, file);
				if (read == 0) { /* this is really error */
					fseek(file, -(offset + 4), SEEK_CUR);
					goto DEFAULT;
				}
				offset += read;
				if (header[0] == 'd' && header[1] == 'a' &&\
				    header[2] == 't' && header[3] == 'a')
					break;
			}
			fread(header, 4, 1, file);
			return ftell(file);
		} else {
			/* this is really error */
			fseek(file, -8, SEEK_CUR);
		}
	} else if (header[0] == 'I' && header[1] == 'D' && header[2] == '3') { 
		rewind(file);
		return 0;
	} else {
		rewind(file);
	}

DEFAULT:
	for (i = 0; i < 8192; i++) {
		fread(header, 4, 1, file);
		fseek(file, -3, SEEK_CUR);
		if (is_valid_header(header))
			break;
	}

	fseek(file, -1, SEEK_CUR);
	return ftell(file);
}

int getID3v2len(FILE *file) {
	int i;
	unsigned char header[10];
	long offset;

	fread(header, sizeof (header), 1, file);

	if (header[0] == 'I' && header[1] == 'D' && header[2] == '3') {
		/*   ID3v2  ã ̴ */
		offset = (unsigned long)(header[6] << 21) |
			 (unsigned long)(header[7] << 14) |
			 (unsigned long)(header[8] <<  7) |
			 (unsigned long)(header[9]);
		/* ID3 ± (offset)ŭ ̵ؾ data frame 
		 *  ȴ
		 * offset header ü ũ ܵǾ ִ */
		fseek(file, offset, SEEK_CUR);
		fread(header, 4, 1, file);
		if (is_valid_header(header)) {
			fseek(file, -4, SEEK_CUR);
			return (int)ftell(file);
		} else {
			for (i = 0; i < 8192; i++) {
				fread(header, 4, 1, file);
				if (is_valid_header(header))
					break;
				fseek(file, -3, SEEK_CUR);
			}
			fseek(file, -4, SEEK_CUR);
			return (int)ftell(file);
		}
	} else {
		fseek(file, -sizeof (header), SEEK_CUR);
		return 0;
	}
}

static int get_mpeg_version_index(int num) {
	switch (num) {
	case 0:
		return 2;	/* version 2.5 */
	case 2:
		return 1;	/* version 2 */
	case 3:
		return 0;	/* version 1 */
	case 1:
	default:
		return 0;	/* reserved, default verion 1 */
	}

}

static int get_mpeg_layer_index(int num) {
	switch (num) {
	case 1:
		return 2;	/* layer III */
	case 2:
		return 1;	/* layer II */
	case 3:
		return 0;	/* layer I */
	case 0:
	default:
		return 0;	/* reserved, default layer III */
	}
}

static int bigendian4(unsigned char *buf) {
	return (((int)buf[0] << 24) | ((int)buf[1] << 16) | \
		((int)buf[2] << 8)  |  (int)buf[3]);
}

static int getxingheader(FILE *file, struct mp3info *info) {
	int id, mode, flags, nframes;
	unsigned char *p;
	unsigned char buf[256];

	info->xing = FALSE;
	fread(buf, sizeof (buf), 1, file);
	fseek(file, -sizeof(buf), SEEK_CUR);

	id = (buf[1] >> 3) & 0x1;
	mode = (buf[3] >> 6) & 0x3;

	p = buf;
	if (id) {
		if (mode == 3)
			p += (17 + 4);
		else
			p += (32 + 4);
	} else {
		if (mode == 3)
			p += (9 + 4);
		else
			p += (17 + 4);
	}

	/* 4 Xing (id)
	 * 4 flags
	 * 4 nframes
	 * 4 size
	 * 100 toc */
	/* check for 'Xing' */
	if (p[0] != 'X' || p[1] != 'i' || p[2] != 'n' || p[3] != 'g')
		return -1;
	p += 4;

	flags = bigendian4(p);
	p += 4;
	if (flags & 0x1) {
		/* xing header has no. of frames */
		nframes = bigendian4(p);
	} else
		return -1;

	info->xing = TRUE;
	info->nframes = nframes;
	info->bitrate = -1;

	return 0;
}

static int getframeheader(FILE *file, struct mp3info *info) {
	unsigned char header[4];	/* this must be 32 bits length */

	fread(&header, sizeof (header), 1, file);

	info->version = get_mpeg_version_index((header[1] >> 3) & 0x3);
	info->layer = get_mpeg_layer_index((header[1] >> 1) & 0x3);
	if (!info->xing)
		info->bitrate = bitrate_tbl[info->version][info->layer][(header[2] >> 4) & 0xF];
	info->frequency = frequency_tbl[info->version][(header[2] >> 2) & 0x3];
	info->padding = (header[2] >> 1) & 0x1;
	info->mode = (header[3] >> 6) & 0x3;

	info->original = (header[3] >> 2) & 0x1;
	
	return 0;
}

static int getID3tag(FILE *file, struct mp3info *info) {
	char buf[3];
	fseek(file, -128L, SEEK_END);

	fread(buf, 3, 1, file);
	if (buf[0] == 'T' && buf[1] == 'A' && buf[2] == 'G') {
		info->id3tag = TRUE;
		fread(info->title, 30, 1, file);
		info->title[30] = '\0';
		fread(info->artist, 30, 1, file);
		info->artist[30] = '\0';
		fread(info->album, 30, 1, file);
		info->album[30] = '\0';
		fread(info->year, 4, 1, file);
		info->year[4] = '\0';
		fread(info->comment, 30, 1, file);
		if (info->comment[28] == 0)
			info->track = info->comment[29];
		else
			info->track = -1;
		info->comment[29] = '\0';
		info->comment[30] = '\0';
		info->genre = getc(file);

		return 128;
	} else {
		info->id3tag = FALSE;
		info->title[0] = '\0';
		info->artist[0] = '\0';
		info->album[0] = '\0';
		info->year[0] = '\0';
		info->comment[0] = '\0';
		info->genre = -1;

		return 0;
	}
}

static long getfilesize(FILE *file) {
	struct stat st;

	if (fstat(fileno(file), &st))
		return 0;
	else
		return st.st_size;
}

int mp3_get_info(const char *filename, struct mp3info *info) {
	int taglen;
	FILE *file;

	file = fopen(filename, "r");
	if (file == NULL)
		return -1;

	taglen = checkheader(file);
	taglen += getID3v2len(file);
	getxingheader(file, info);
	getframeheader(file, info);
	taglen += getID3tag(file, info);
	info->size = getfilesize(file);
	/* second formula */
	/* timelength (sec) = total_no_of_frame / frame_per_second 
		tatal_no_of_frame = data_size / frame_size
		data_size = file_size - tag_length
		frame_size = 144 * bitrate / (frequency + padding)
		frame_per_second = 1000 / 26 = 38
			1 frame lasts for 26 milisecond
	*/
	if (info->xing)
		info->seconds = info->nframes * 26 / 1000 + 1;
	else
		info->seconds = (int)((double)(info->size - taglen) / 1000.0 \
				* (double)(info->frequency + info->padding) \
				/ 1000.0 * 26 \
				/ 144.0 / ((double)info->bitrate)) \
				+ 1; /* for ceiling */

	fclose(file);
	return 0;
}

int mp3_write_id3tag(const char *filename, const char *title, \
					   const char *artist, \
					   const char *album, \
					   const char *comment, \
					   const char *year, \
					   int track, \
					   int genre) {
	char buf[128];
	FILE *file;

	file = fopen(filename, "r+");
	if (file == NULL)
		return -1;

	fseek(file, -128L, SEEK_END); 
	fread(buf, sizeof(buf), 1, file);

	if (buf[0] == 'T' && buf[1] == 'A' && buf[2] == 'G')
		fseek(file, -128L, SEEK_END); 
	else
		fseek(file, 0L, SEEK_END); 

	memset(buf, ' ', sizeof(buf));
	buf[0] = 'T';
	buf[1] = 'A';
	buf[2] = 'G';
	copy_string(&buf[3],  title,   30);
	copy_string(&buf[33], artist,  30);
	copy_string(&buf[63], album,   30);
	copy_string(&buf[93], year,     4);
	copy_string(&buf[97], comment, 30);
	if (track > 0) {
		buf[125] = 0;
		buf[126] = (char)track;
	}
	buf[127] = (char)genre;

	fwrite(buf, sizeof(buf), 1, file);
	fclose(file);

	return 0;
}

int mp3_remove_id3tag(const char *filename) {
	char buf[128];
	long len;
	FILE *file;

	file = fopen(filename, "r+");
	if (file == NULL)
		return -1;

	fseek(file, -128L, SEEK_END); 
	len = ftell(file);
	fread(buf, sizeof(buf), 1, file);

	if (buf[0] == 'T' && buf[1] == 'A' && buf[2] == 'G') 
		return ftruncate(fileno(file), len);

	return 0;
}
